@@ -45,21 +45,27 @@ func deferUnlambda(m dsl.Matcher) {
 //doc:after   io.ReadAll(r)
 func ioutilDeprecated(m dsl.Matcher) {
 	m.Match(`ioutil.ReadAll($_)`).
+		Where(m.GoVersion().GreaterEqThan("1.16")).
 		Report(`ioutil.ReadAll is deprecated, use io.ReadAll instead`)
 
 	m.Match(`ioutil.ReadFile($_)`).
+		Where(m.GoVersion().GreaterEqThan("1.16")).
 		Report(`ioutil.ReadFile is deprecated, use os.ReadFile instead`)
 
 	m.Match(`ioutil.WriteFile($_, $_, $_)`).
+		Where(m.GoVersion().GreaterEqThan("1.16")).
 		Report(`ioutil.WriteFile is deprecated, use os.WriteFile instead`)
 
 	m.Match(`ioutil.ReadDir($_)`).
+		Where(m.GoVersion().GreaterEqThan("1.16")).
 		Report(`ioutil.ReadDir is deprecated, use os.ReadDir instead`)
 
 	m.Match(`ioutil.NopCloser($_)`).
+		Where(m.GoVersion().GreaterEqThan("1.16")).
 		Report(`ioutil.NopCloser is deprecated, use io.NopCloser instead`)
 
 	m.Match(`ioutil.Discard`).
+		Where(m.GoVersion().GreaterEqThan("1.16")).
 		Report(`ioutil.Discard is deprecated, use io.Discard instead`)
 }
 
@@ -112,12 +118,14 @@ func httpNoBody(m dsl.Matcher) {
 	m.Match("http.NewRequest($method, $url, $nil)").
 		Where(m["nil"].Text == "nil").
 		Suggest("http.NewRequest($method, $url, http.NoBody)").
-		Report("http.NoBody should be preferred to the nil request body")
+		Report("http.NoBody should be preferred to the nil request body").
+		At(m["nil"])
 
 	m.Match("http.NewRequestWithContext($ctx, $method, $url, $nil)").
 		Where(m["nil"].Text == "nil").
 		Suggest("http.NewRequestWithContext($ctx, $method, $url, http.NoBody)").
-		Report("http.NoBody should be preferred to the nil request body")
+		Report("http.NoBody should be preferred to the nil request body").
+		At(m["nil"])
 }
 
 //doc:summary Detects expressions like []rune(s)[0] that may cause unwanted rune slice allocation
@@ -323,17 +331,20 @@ func assignOp(m dsl.Matcher) {
 	m.Match(`$x = $x &^ $y`).Where(m["x"].Pure).Report("replace `$$` with `$x &^= $y`")
 }
 
-//doc:summary Detects WriteRune calls with byte literal argument and reports to use WriteByte instead
-//doc:tags    performance experimental
+//doc:summary Detects WriteRune calls with rune literal argument that is single byte and reports to use WriteByte instead
+//doc:tags    performance experimental opinionated
 //doc:before  w.WriteRune('\n')
 //doc:after   w.WriteByte('\n')
 func preferWriteByte(m dsl.Matcher) {
+	// utf8.RuneSelf:
+	// characters below RuneSelf are represented as themselves in a single byte.
+	const runeSelf = 0x80
 	m.Match(`$w.WriteRune($c)`).Where(
-		m["w"].Type.Implements("io.ByteWriter") && (m["c"].Const && m["c"].Value.Int() < 256),
-	).Report(`consider replacing $$ with $w.WriteByte($c)`)
+		m["w"].Type.Implements("io.ByteWriter") && (m["c"].Const && m["c"].Value.Int() < runeSelf),
+	).Report(`consider writing single byte rune $c with $w.WriteByte($c)`)
 }
 
-//doc:summary Detects fmt.Sprint(f|ln) calls which can be replaced with fmt.Fprint(f|ln)
+//doc:summary Detects fmt.Sprint(f/ln) calls which can be replaced with fmt.Fprint(f/ln)
 //doc:tags    performance experimental
 //doc:before  w.Write([]byte(fmt.Sprintf("%x", 10)))
 //doc:after   fmt.Fprintf(w, "%x", 10)
@@ -352,6 +363,10 @@ func preferFprint(m dsl.Matcher) {
 		Where(m["w"].Type.Implements("io.Writer")).
 		Suggest("fmt.Fprintln($w, $args)").
 		Report(`fmt.Fprintln($w, $args) should be preferred to the $$`)
+
+	m.Match(`io.WriteString($w, fmt.Sprint($*args))`).Suggest(`fmt.Fprint($w, $args)`)
+	m.Match(`io.WriteString($w, fmt.Sprintf($*args))`).Suggest(`fmt.Fprintf($w, $args)`)
+	m.Match(`io.WriteString($w, fmt.Sprintln($*args))`).Suggest(`fmt.Fprintln($w, $args)`)
 }
 
 //doc:summary Detects suspicious duplicated arguments
@@ -404,8 +419,8 @@ func dupArg(m dsl.Matcher) {
 
 //doc:summary Detects suspicious http.Error call without following return
 //doc:tags    diagnostic experimental
-//doc:before  x + string(os.PathSeparator) + y
-//doc:after   filepath.Join(x, y)
+//doc:before  if err != nil { http.Error(...); }
+//doc:after   if err != nil { http.Error(...); return; }
 func returnAfterHttpError(m dsl.Matcher) {
 	m.Match(`if $_ { $*_; http.Error($w, $err, $code) }`).
 		Report("Possibly return is missed after the http.Error call").
@@ -558,7 +573,7 @@ func equalFold(m dsl.Matcher) {
 		`strings.ToUpper($x) == strings.ToUpper($y)`,
 		`$x == strings.ToUpper($y)`).
 		Where(m["x"].Pure && m["y"].Pure && m["x"].Text != m["y"].Text).
-		Suggest(`strings.EqualFold($x, $y)]`).
+		Suggest(`strings.EqualFold($x, $y)`).
 		Report(`consider replacing with strings.EqualFold($x, $y)`)
 
 	// string != patterns
@@ -570,7 +585,7 @@ func equalFold(m dsl.Matcher) {
 		`strings.ToUpper($x) != strings.ToUpper($y)`,
 		`$x != strings.ToUpper($y)`).
 		Where(m["x"].Pure && m["y"].Pure && m["x"].Text != m["y"].Text).
-		Suggest(`!strings.EqualFold($x, $y)]`).
+		Suggest(`!strings.EqualFold($x, $y)`).
 		Report(`consider replacing with !strings.EqualFold($x, $y)`)
 
 	// bytes.Equal patterns
@@ -582,7 +597,7 @@ func equalFold(m dsl.Matcher) {
 		`bytes.Equal(bytes.ToUpper($x), bytes.ToUpper($y))`,
 		`bytes.Equal($x, bytes.ToUpper($y))`).
 		Where(m["x"].Pure && m["y"].Pure && m["x"].Text != m["y"].Text).
-		Suggest(`bytes.EqualFold($x, $y)]`).
+		Suggest(`bytes.EqualFold($x, $y)`).
 		Report(`consider replacing with bytes.EqualFold($x, $y)`)
 }
 
@@ -625,15 +640,17 @@ func stringConcatSimplify(m dsl.Matcher) {
 //doc:before  t.Unix() / 1000
 //doc:after   t.UnixMilli()
 func timeExprSimplify(m dsl.Matcher) {
+	isTime := func(v dsl.Var) bool {
+		return v.Type.Is(`time.Time`) || v.Type.Is(`*time.Time`)
+	}
+
 	m.Match(`$t.Unix() / 1000`).
-		Where(m.GoVersion().GreaterEqThan("1.17") &&
-			(m["t"].Type.Is(`time.Time`) || m["t"].Type.Is(`*time.Time`))).
+		Where(m.GoVersion().GreaterEqThan("1.17") && isTime(m["t"])).
 		Suggest("$t.UnixMilli()").
 		Report(`use $t.UnixMilli() instead of $$`)
 
 	m.Match(`$t.UnixNano() * 1000`).
-		Where(m.GoVersion().GreaterEqThan("1.17") &&
-			(m["t"].Type.Is(`time.Time`) || m["t"].Type.Is(`*time.Time`))).
+		Where(m.GoVersion().GreaterEqThan("1.17") && isTime(m["t"])).
 		Suggest("$t.UnixMicro()").
 		Report(`use $t.UnixMicro() instead of $$`)
 }
@@ -643,20 +660,24 @@ func timeExprSimplify(m dsl.Matcher) {
 //doc:before  type Foo struct{ ...; sync.Mutex; ... }
 //doc:after   type Foo struct{ ...; mu sync.Mutex; ... }
 func exposedSyncMutex(m dsl.Matcher) {
+	isExported := func(v dsl.Var) bool {
+		return v.Text.Matches(`^\p{Lu}`)
+	}
+
 	m.Match(`type $x struct { $*_; sync.Mutex; $*_ }`).
-		Where(m["x"].Text.Matches(`^\p{Lu}`)).
+		Where(isExported(m["x"])).
 		Report("don't embed sync.Mutex")
 
 	m.Match(`type $x struct { $*_; *sync.Mutex; $*_ }`).
-		Where(m["x"].Text.Matches(`^\p{Lu}`)).
+		Where(isExported(m["x"])).
 		Report("don't embed *sync.Mutex")
 
 	m.Match(`type $x struct { $*_; sync.RWMutex; $*_ }`).
-		Where(m["x"].Text.Matches(`^\p{Lu}`)).
+		Where(isExported(m["x"])).
 		Report("don't embed sync.RWMutex")
 
 	m.Match(`type $x struct { $*_; *sync.RWMutex; $*_ }`).
-		Where(m["x"].Text.Matches(`^\p{Lu}`)).
+		Where(isExported(m["x"])).
 		Report("don't embed *sync.RWMutex")
 }
 
@@ -700,3 +721,18 @@ func emptyDecl(m dsl.Matcher) {
 	m.Match(`const()`).Report(`empty const() block`)
 	m.Match(`type()`).Report(`empty type() block`)
 }
+
+//doc:summary Detects suspicious formatting strings usage
+//doc:tags    diagnostic experimental
+//doc:before  fmt.Errorf(msg)
+//doc:after   fmt.Errorf("%s", msg)
+func dynamicFmtString(m dsl.Matcher) {
+	m.Match(`fmt.Errorf($f)`).
+		Where(!m["f"].Const).
+		Suggest("errors.New($f)").
+		Report(`use errors.New($f) or fmt.Errorf("%s", $f) instead`)
+
+	m.Match(`fmt.Errorf($f($*args))`).
+		Suggest("errors.New($f($*args))").
+		Report(`use errors.New($f($*args)) or fmt.Errorf("%s", $f($*args)) instead`)
+}
